# -*- coding: utf-8 -*-
"""Various utilities for dealing with syntax trees.

Include base classes to implement a visitor pattern.

"""
# utils pending deprecation
from . import visitor
from .deprecated_utils import (  # noqa: F401
    LuceneTreeTransformer, LuceneTreeVisitor, LuceneTreeVisitorV2)
from .tree import AndOperation, BaseOperation, OrOperation, BoolOperation


class UnknownOperationResolver(visitor.TreeTransformer):
    """Transform the UnknownOperation to OR or AND
    """

    VALID_OPERATIONS = frozenset([None, AndOperation, OrOperation, BoolOperation])
    DEFAULT_OPERATION = AndOperation

    def __init__(self, resolve_to=None, add_head=" "):
        """Initialize a new resolver

        :param resolve_to: must be either None, OrOperation, AndOperation, BoolOperation.

          for the latter three the UnknownOperation is replaced by specified operation.

          if it is None, we use the last operation encountered, as would Lucene do
        """
        if resolve_to not in self.VALID_OPERATIONS:
            raise ValueError("%r is not a valid value for resolve_to" % resolve_to)
        self.resolve_to = resolve_to
        self.add_head = add_head
        super().__init__(track_parents=True)

    def _last_operation(self, context):
        return context.setdefault("last_operation", {})

    def _first_nonop_parent(self, parents):
        for parent in parents:
            if not isinstance(parent, BaseOperation):
                return id(parent)  # use id() because parent might not be hashable
        else:
            return None

    def _track_last_op(self, node, context):
        if self.resolve_to is None:
            # next unknow operation at same level should resolve to my type of operation
            # so track it
            parent = self._first_nonop_parent(context.get("parents", []))
            self._last_operation(context)[parent] = type(node)

    def _get_last_op(self, node, context):
        # search for last operation
        parent = self._first_nonop_parent(context.get("parents", []))
        return self._last_operation(context).get(parent, self.DEFAULT_OPERATION)

    def visit_or_operation(self, node, context):
        self._track_last_op(node, context)
        yield from self.generic_visit(node, context)

    def visit_and_operation(self, node, context):
        self._track_last_op(node, context)
        yield from self.generic_visit(node, context)

    def visit_unknown_operation(self, node, context):
        # resolve
        if self.resolve_to is None:
            operation = self._get_last_op(node, context)
        else:
            operation = self.resolve_to
        new_node = operation(pos=node.pos, size=node.size, head=node.head, tail=node.tail)
        new_node.children = self.clone_children(node, new_node, context)
        # add head to children but the first to separate element from operation (x y --> x AND y)
        for child in new_node.children[1:]:
            child.head = self.add_head + child.head
        yield new_node

    def __call__(self, tree):
        return self.visit(tree)


def normalize_nested_fields_specs(nested_fields):
    """normalize nested_fields specification to only have nested dicts

    :param dict nested_fields:  dict contains fields that are nested in ES
        each nested fields contains either
        a dict of nested fields
        (if some of them are also nested)
        or a list of nesdted fields (this is for commodity)


    ::
        >>> from unittest import TestCase
        >>> TestCase().assertDictEqual(
        ...     normalize_nested_fields_specs(
        ...         {"author" : {"books": ["name", "ref"], "firstname" : None }}),
        ...     {"author" : {"books": {"name": {}, "ref": {}}, "firstname" : {} }})
    """
    if nested_fields is None:
        return {}
    elif isinstance(nested_fields, dict):
        return {k: normalize_nested_fields_specs(v) for k, v in nested_fields.items()}
    else:
        # should be an iterable, transform to dict
        return {sub: {} for sub in nested_fields}


def _flatten_fields_specs(object_fields):
    if not object_fields:
        return [[]]  # parent is a single field
    elif isinstance(object_fields, dict):
        return [
            [k] + v2
            for k, v in object_fields.items()
            for v2 in _flatten_fields_specs(v)]
    else:  # iterable
        return [[k] for k in object_fields]


def flatten_nested_fields_specs(nested_fields):
    """normalize object_fields specification to only have a simple set

    :param dict nested_fields:  contains fields that are object in ES
        has a serie of nested dict.
        List are accepted as well for concisness.

    ::
        >>> from unittest import TestCase
        >>> flatten_nested_fields_specs(None)
        set()
        >>> TestCase().assertEqual(
        ...     flatten_nested_fields_specs(["author.name", "book.title"]),
        ...     set(["author.name", "book.title"]))
        >>> TestCase().assertEqual(
        ...     flatten_nested_fields_specs(
        ...         {"book" : { "author": ["firstname", "lastname"], "title" : None }}),
        ...     set(["book.author.firstname", "book.author.lastname", "book.title"]))
    """
    if isinstance(nested_fields, dict):
        return set(".".join(v) for v in _flatten_fields_specs(nested_fields))
    elif nested_fields is None:
        return set([])
    else:
        return set(nested_fields)


def normalize_object_fields_specs(object_fields):
    """normalize object_fields specification to only have a simple set

    :param dict object_fields:  contains fields that are object in ES
        has a serie of nested dict.
        List are accepted as well for concisness.
        None, which means no spec, is returned as is.

    ::
        >>> from unittest import TestCase
        >>> normalize_object_fields_specs(None) is None
        True
        >>> TestCase().assertEqual(
        ...     normalize_object_fields_specs(["author.name", "book.title"]),
        ...     set(["author.name", "book.title"]))
        >>> TestCase().assertEqual(
        ...     normalize_object_fields_specs(
        ...         {"book" : { "author": ["firstname", "lastname"], "title" : None }}),
        ...     set(["book.author.firstname", "book.author.lastname", "book.title"]))
    """
    if object_fields is None:
        return None
    if isinstance(object_fields, dict):
        return set(".".join(v) for v in _flatten_fields_specs(object_fields))
    else:
        return set(object_fields)



"""Utilities to manage head and tail of elements

The scope is to avoid loosing part of the original text in the final tree.
"""
from .tree import Item


class TokenValue:

    def __init__(self, value):
        self.value = value
        self.pos = None
        self.size = None
        self.head = ""
        self.tail = ""

    def __repr__(self):
        return "TokenValue(%s)" % self.value

    def __str__(self):
        return str(self.value) if self.value else ""


class HeadTailLexer:
    """Utility to handle head and tail at lexer time.
    """

    LEXER_ATTR = "_luqum_headtail"

    @classmethod
    def handle(cls, token, orig_value):
        """Handling a token.

        .. note::
          PLY does not gives acces to previous tokens,
          although it does not provide any infrastructure for handling specific state.

          So we use the strategy
          of puting a :py:cls:`HeadTailLexer`instance as an attribute of the lexer
          each time we start a new tokenization.
        """
        # get instance
        if token.lexpos == 0:
            # first token make instance
            instance = cls()
            setattr(token.lexer, cls.LEXER_ATTR, instance)
        else:
            instance = getattr(token.lexer, cls.LEXER_ATTR)
        # handle
        instance.handle_token(token, orig_value)

    def __init__(self):
        self.head = None
        """This will track the head of next element, useful only for first element
        """
        self.last_elt = None
        """This will track the last token, so we can use it to add the tail to it.
        """

    def handle_token(self, token, orig_value):
        """Handle head and tail for tokens

        The scope is to avoid loosing part of the original text and keep it in elements.
        """
        # handle headtail
        if token.type == "SEPARATOR":
            if token.lexpos == 0:
                # spaces at expression start, head for next token
                self.head = token.value
            else:
                # tail of last processed token
                if self.last_elt is not None:
                    self.last_elt.value.tail += token.value
        else:
            # if there is a head, apply
            head = self.head
            if head is not None:
                token.value.head = head
                self.head = None
            # keep tracks of token, to apply tail later
            self.last_elt = token
        # also set pos and size
        if isinstance(token.value, (Item, TokenValue)):
            token.value.pos = token.lexpos
            token.value.size = len(orig_value)


token_headtail = HeadTailLexer.handle


class HeadTailManager:
    """Utility to hande head and tail at expression parse time
    """

    def pos(self, p, head_transfer=False, tail_transfer=False):
        """Compute pos and size of element 0 based on it's parts (p[1:])

        :param list p: the parser expression as in PLY
        :param bool head_transfer: True if head of first child will be transfered to p[0]
        :param bool tail_transfer: True if tail of last child wiil be transfered to p[0]
        """
        # pos
        if p[1].pos is not None:
            p[0].pos = p[1].pos
            if not head_transfer:
                # head is'nt transfered, so we are before it
                p[0].pos -= len(p[1].head)
        # size
        p[0].size = sum(
            (elt.size or 0) + len(elt.head or "") + len(elt.tail or "") for elt in p[1:])
        if head_transfer and p[1].head:
            # we account head in size, remove it
            p[0].size -= len(p[1].head)
        last_p = p[len(p) - 1]  # negative indexing not supported by PLY
        if tail_transfer and last_p.tail:
            # we account head in size, remove it
            p[0].size -= len(last_p.tail)

    def binary_operation(self, p, op_tail):
        self.pos(p, head_transfer=False, tail_transfer=False)
        # correct size
        p[0].size -= len(op_tail)

    def simple_term(self, p):
        self.pos(p, head_transfer=True, tail_transfer=True)
        p[0].head = p[1].head
        p[0].tail = p[1].tail

    def unary(self, p):
        """OP expr"""
        self.pos(p, head_transfer=True, tail_transfer=False)
        p[0].head = p[1].head
        p[2].head = p[1].tail + p[2].head

    def post_unary(self, p):
        """expr OP"""
        self.pos(p, head_transfer=False, tail_transfer=True)
        p[1].tail += p[2].head
        p[0].tail = p[2].tail

    def paren(self, p):
        """( expr )"""
        self.pos(p, head_transfer=True, tail_transfer=True)
        # p[0] is global element (Group or FieldGroup)
        # p[2] is content
        # p[1] is left parenthesis
        p[0].head = p[1].head
        p[2].head = p[1].tail + p[2].head
        # p[3] is right parenthesis
        p[2].tail += p[3].head
        p[0].tail = p[3].tail

    def range(self, p):
        """[ expr TO expr ]"""
        self.pos(p, head_transfer=True, tail_transfer=True)
        # p[0] is global element (Range)
        # p[2] is lower bound
        p[0].head = p[1].head
        p[2].head = p[1].tail + p[2].head
        # p[3] is TO
        # p[4] is upper bound
        p[2].tail += p[3].head
        p[4].head = p[3].tail + p[4].head
        # p[5] is upper braket
        p[4].tail += p[5].head
        p[0].tail = p[5].tail

    def search_field(self, p):
        """name: expr"""
        self.pos(p, head_transfer=True, tail_transfer=False)
        # p[0] is global element (SearchField)
        # p[1] is search field name
        # p[2] is COLUMN
        p[0].head = p[1].head
        if p[1].tail or p[2].head:
            pass  # FIXME: add warning, or handle space between point and name in SearchField ?
        # p[3] is the expression
        p[3].head = p[2].tail + p[3].head


head_tail = HeadTailManager()
"""singleton of HeadTailManager
"""



"""It can be teadious to add spaces in a tree you generate programatically.

This module provide a utility to transform a tree so that it contains necessary head/tail
for expression to be printable.
"""

from . import visitor


class AutoHeadTail(visitor.TreeTransformer):
    """This class implements a transformer so that hand built tree,
    can have reasonable values for `head` and `tail` on their items,
    in order for the expression to be printable.
    """

    SPACER = " "

    def add_head(self, node):
        if not node.head:
            node.head = self.SPACER

    def add_tail(self, node):
        if not node.tail:
            node.tail = self.SPACER

    def visit_base_operation(self, node, context):
        new_node = node.clone_item()
        children = list(self.clone_children(node, new_node, context))
        # add tail to first node
        self.add_tail(children[0])
        # add head and tail to inner nodes
        for child in children[1:-1]:
            self.add_head(child)
            self.add_tail(child)
        # add head to last
        self.add_head(children[-1])
        new_node.children = children
        yield new_node

    def visit_unknown_operation(self, node, context):
        new_node = node.clone_item()
        children = list(self.clone_children(node, new_node, context))
        # add tail to each node, but last
        for child in children[:-1]:
            self.add_tail(child)
        new_node.children = children
        yield new_node

    def visit_not(self, node, context):
        new_node = node.clone_item()
        children = list(self.clone_children(node, new_node, context))
        # add head to children, to have space between NOT and sub expression
        self.add_head(children[0])
        new_node.children = children
        yield new_node

    def visit_range(self, node, context):
        new_node = node.clone_item()
        children = list(self.clone_children(node, new_node, context))
        # add tail to lower_bound, and head to upper bound
        self.add_tail(children[0])
        self.add_head(children[-1])
        new_node.children = children
        yield new_node

    def __call__(self, tree):
        new_tree = self.visit(tree)
        return new_tree


auto_head_tail = AutoHeadTail()
"""method to auto add head and tail to items of a lucene tree so that it is printable
"""



# -*- coding: utf-8 -*-
"""Elements that will constitute the parse tree of a query.

You may use these items to build a tree representing a query,
or get a tree as the result of parsing a query string.
"""
import re
from decimal import Decimal

_MARKER = object()


class Item(object):
    """Base class for all items that compose the parse tree.

    An item is a part of a request.

    :param int pos: position of element in orginal text (not accounting for head)
    :param int size: size of element in orginal text (not accounting for head and tail)
    :param str head: non meaningful text before this element
    :param str tail: non meaningful text after this element
    """

    # /!\ Note on Item (and subclasses) __magic__ methods: /!\
    #
    # Since we're dealing with recursive structures, we must avoid using
    # the builtin helper methods when dealing with nested objects in
    # __magic__ methods.
    #
    # As the helper usually calls the relevant method, we end up with two
    # function calls instead of one, and end up hitting python's max recursion
    # limit twice as fast!
    #
    # This is why we're calling c.__repr__ instead of repr(c) in the __repr__
    # method. Same thing applies for all magic methods (__str__, __eq__, and any
    # other we might add in the future).

    #: this attribute permits to list attributes that participate in telling equality of two item
    #: this excludes children (for generic `__eq__` methode will already recursively compare them)
    _equality_attrs = []
    #: this attribute permits to list attributes that defines the children.
    #: Order is important.
    _children_attrs = []

    def __init__(self, pos=None, size=None, head="", tail=""):
        self.pos = pos
        self.size = size
        self.head = head
        self.tail = tail

    def clone_item(self, **kwargs):
        """clone an item, but not its children !

        This is particularly useful for the :py:class:`.visitor.TreeTransformer` pattern.

        :param dict kwargs: those item will be added to `__init__` call.
            It's a simple way to change some values of target item.
        """
        return self._clone_item(cls=type(self), **kwargs)

    def _clone_item(self, cls, *args, **kwargs):
        """internal implementation of clone_item (for specific sub classes tweaks)

        :param type cls: the new class
        """
        attrs = {"pos": self.pos, "size": self.size, "head": self.head, "tail": self.tail}
        # we can _equality_attrs as they normally correspond to what we need to copy
        attrs.update((attr_name, getattr(self, attr_name)) for attr_name in self._equality_attrs)
        # use NoneItem for children using _children_attrs
        attrs.update((attr_name, NONE_ITEM) for attr_name in self._children_attrs)
        # add optional kwargs
        attrs.update(**kwargs)
        return cls(*args, **attrs)

    @property
    def children(self):
        """As base of a tree structure, an item may have children"""
        return [getattr(self, attr) for attr in self._children_attrs]

    @children.setter
    def children(self, value):
        """generic setter for children.

        Having a setter for children in useful for generic manipulations
        like in :py:mod:`luqum.visitor`
        """
        if len(value) != len(self._children_attrs):
            num_children = len(value) if value else "no"
            raise ValueError(
                f"{type(self)} accepts {num_children} children,"
                f" and you try to set {len(value)} children"
            )
        for attr, v in zip(self._children_attrs, value):
            setattr(self, attr, v)

    def _head_tail(self, value, head_tail):
        if head_tail:
            return self.head + value + self.tail
        else:
            return value

    def span(self, head_tail=False):
        """return (start, end) position of this element in global expression.

        :param bool head_tail: should span include head and tail of element ?
        """
        if self.pos is None:
            start, end = None, None
        else:
            start = self.pos - (len(self.head) if head_tail else 0)
            end = self.pos + self.size + (len(self.tail) if head_tail else 0)
        return start, end

    def __repr__(self):
        children = ", ".join(c.__repr__() for c in self.children)
        return "%s(%s)" % (self.__class__.__name__, children)

    def __eq__(self, other):
        """a generic equal operation

        It make uses of :py:attr:`Item._equality_attrs`,
        and also recursively compare children
        """
        return (
            self is other  # shortcut
        ) or (
            self.__class__ == other.__class__ and
            len(self.children) == len(other.children) and
            all(getattr(self, a, _MARKER) == getattr(other, a, _MARKER)
                for a in self._equality_attrs) and
            all(c.__eq__(d) for c, d in zip(self.children, other.children))
        )


class NoneItem(Item):
    """This Item is a place holder, think to it as None.

    It can be used, eg. to initialize an element childrens, until we feed in the real children.
    """

    def __str__(self, head_tail=False):
        return ""


#: an instanciation of NoneItem, as it is always the same
NONE_ITEM = NoneItem()


class SearchField(Item):
    """Indicate wich field the search expression operates on

    eg: *desc* in ``desc:(this OR that)``

    :param str name: name of the field
    :param expr: the searched expression
    """
    _equality_attrs = ['name']
    _children_attrs = ["expr"]

    def __init__(self, name, expr, **kwargs):
        self.name = name
        self.expr = expr
        super().__init__(**kwargs)

    def __str__(self, head_tail=False):
        value = self.name + ":" + self.expr.__str__(head_tail=True)
        return self._head_tail(value, head_tail)

    def __repr__(self):
        return "SearchField(%r, %s)" % (self.name, self.expr.__repr__())


class BaseGroup(Item):
    """Base class for group of expressions or field values

    :param expr: the expression inside parenthesis
    """
    _children_attrs = ["expr"]

    def __init__(self, expr, **kwargs):
        self.expr = expr
        super().__init__(**kwargs)

    def __str__(self, head_tail=False):
        value = "(%s)" % self.expr.__str__(head_tail=True)
        return self._head_tail(value, head_tail)


class Group(BaseGroup):
    """Group sub expressions
    """


class FieldGroup(BaseGroup):
    """Group values for a query on a field
    """


def group_to_fieldgroup(g):
    return FieldGroup(g.expr, pos=g.pos, size=g.size, head=g.head, tail=g.tail)


class Range(Item):
    """A Range

    :param low: lower bound
    :param high: higher bound
    :param bool include_low: wether lower bound is included
    :param bool include_high: wether higher bound is included
    """

    LOW_CHAR = {True: '[', False: '{'}
    HIGH_CHAR = {True: ']', False: '}'}

    _equality_attrs = ['include_high', 'include_low']
    _children_attrs = ["low", "high"]

    def __init__(self, low, high, include_low=True, include_high=True, **kwargs):
        self.low = low
        self.high = high
        self.include_low = include_low
        self.include_high = include_high
        super().__init__(**kwargs)

    def __str__(self, head_tail=False):
        value = "%s%sTO%s%s" % (
            self.LOW_CHAR[self.include_low],
            self.low.__str__(head_tail=True),
            self.high.__str__(head_tail=True),
            self.HIGH_CHAR[self.include_high])
        return self._head_tail(value, head_tail)


class Term(Item):
    """Base for terms

    :param str value: the value
    """
    WILDCARDS_PATTERN = re.compile(r"((?<=[^\\])[?*]|\\\\[?*]|^[?*])")  # non escaped * and ?
    # see
    # https://lucene.apache.org/core/3_6_0/queryparsersyntax.html#Escaping%20Special%20Characters
    WORD_ESCAPED_CHARS = re.compile(r'\\([+\-&|!(){}[\]^"~*?:\\])')

    _equality_attrs = ['value']

    def __init__(self, value, **kwargs):
        self.value = value
        super().__init__(**kwargs)

    @property
    def unescaped_value(self):
        # remove '\' that escape characters
        return self.WORD_ESCAPED_CHARS.sub(r'\1', self.value)

    def is_wildcard(self):
        """:return bool: True if value is the wildcard ``*``
        """
        return self.value == "*"

    def iter_wildcards(self):
        """list wildcards contained in value and their positions
        """
        for matched in self.WILDCARDS_PATTERN.finditer(self.value):
            yield matched.span(), matched.group()

    def split_wildcards(self):
        """split term on wildcards
        """
        return self.WILDCARDS_PATTERN.split(self.value)

    def has_wildcard(self):
        """:return bool: True if value contains a wildcards
        """
        return any(self.iter_wildcards())

    def __str__(self, head_tail=False):
        value = self.value
        return self._head_tail(value, head_tail)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, str(self))


class Word(Term):
    """A single word term

    :param str value: the value
    """


class Phrase(Term):
    """A phrase term, that is a sequence of words enclose in quotes

    :param str value: the value, including the quotes. Eg. ``'"my phrase"'``
    """
    def __init__(self, value, **kwargs):
        super(Phrase, self).__init__(value, **kwargs)
        assert self.value.endswith('"') and self.value.startswith('"'), (
               "Phrase value must contain the quotes")


class Regex(Term):
    """A regex term, that is a sequence of words enclose in slashes

    :param str value: the value, including the slashes. Eg. ``'/my regex/'``
    """
    def __init__(self, value, **kwargs):
        super(Regex, self).__init__(value, **kwargs)
        assert value.endswith('/') and value.startswith('/'), (
               "Regex value must contain the slashes")


class BaseApprox(Item):
    """Base for approximations, that is fuzziness and proximity
    """
    _equality_attrs = ['degree']
    _children_attrs = ["term"]

    def __init__(self, term, degree=None, **kwargs):
        self.term = term
        self._implicit_degree = degree is None  # this is just for display
        self.degree = self._normalize_degree(degree)
        super().__init__(**kwargs)

    def __repr__(self):
        return "%s(%s, %s)" % (self.__class__.__name__, self.term.__repr__(), self.degree)

    def __str__(self, head_tail=False):
        value = "%s~%s" % (
            self.term.__str__(head_tail=True),
            self.degree if not self._implicit_degree else "",
        )
        return self._head_tail(value, head_tail)


class Fuzzy(BaseApprox):
    """Fuzzy search on word

    :param Word term: the approximated term
    :param degree: the degree which will be converted to :py:class:`decimal.Decimal`.
    """
    def _normalize_degree(self, degree):
        if degree is None:
            degree = 0.5
        if not isinstance(degree, Decimal):
            degree = Decimal(degree).normalize()
        return degree


class Proximity(BaseApprox):
    """Proximity search on phrase

    :param Phrase term: the approximated phrase
    :param degree: the degree which will be converted to :py:func:`int`.
    """

    def _normalize_degree(self, degree):
        if degree is None:
            degree = 1
        return int(degree)


class Boost(Item):
    """A term for boosting a value or a group there of

    :param expr: the boosted expression
    :param force: boosting force, will be converted to :py:class:`decimal.Decimal`
    """
    _equality_attrs = ['force']
    _children_attrs = ["expr"]

    def __init__(self, expr, force, **kwargs):
        self.expr = expr
        self.force = Decimal(force).normalize() if force is not None else 1
        self.implicit_force = force is None
        super().__init__(**kwargs)

    def __repr__(self):
        return "%s(%s, %s)" % (self.__class__.__name__, self.expr.__repr__(), self.force)

    def __str__(self, head_tail=False):
        force = "" if self.implicit_force else self.force
        value = "%s^%s" % (self.expr.__str__(head_tail=True), force)
        return self._head_tail(value, head_tail)


class BaseOperation(Item):
    """
    Parent class for binary operations are binary operation used to join expressions,
    like OR and AND

    :param operands: expressions to apply operation on
    """

    def __init__(self, *operands, **kwargs):
        self.operands = operands
        super().__init__(**kwargs)

    def __str__(self, head_tail=False):
        value = ("%s" % self.op).join(o.__str__(head_tail=True) for o in self.operands)
        return self._head_tail(value, head_tail)

    @property
    def children(self):
        """children are left and right expressions
        """
        return self.operands

    @children.setter
    def children(self, value):
        """Generic setter for children

        :param iterable value: operands
        """
        self.operands = tuple(value)


class BoolOperation(BaseOperation):
    """Lucene Boolean Query.

    This operation assumes that the query builder can utilize a boolean operator
    with three possible sections, must, should and must_not. If the
    UnknownOperationResolver is asked to resolve_to this operation, the query
    builder can utilize this operator directly instead of nested AND/OR.
    This also makes it possible to correctly support Lucene queries such as:
    "apples +bananas -vegetables".

    .. seealso::
        the :py:class:`.utils.UnknownOperationResolver`
    """
    op = ""


class UnknownOperation(BaseOperation):
    """Unknown Boolean operator.

    .. warning::
        This is used to represent implicit operations (ie: term:foo term:bar),
        as we cannot know for sure which operator should be used.

        Lucene seem to use whatever operator was used before reaching that one,
        defaulting to AND, but we cannot know anything about this at parsing
        time...
    .. seealso::
        the :py:class:`.utils.UnknownOperationResolver` to resolve those nodes to OR and AND
    """
    op = ''


class OrOperation(BaseOperation):
    """OR expression
    """
    op = 'OR'


class AndOperation(BaseOperation):
    """AND expression
    """
    op = 'AND'


def create_operation(cls, a, b, op_tail=" "):
    """Create operation between a and b, merging if a or b is already an operation of same class

    :param a: left operand
    :param b: right operand
    :param op_tail: tail of operation token
    """
    operands = []
    operands.extend(a.operands if isinstance(a, cls) else [a])
    left_operands = b.operands if isinstance(b, cls) else [b]
    left_operands[0].head += op_tail
    operands.extend(left_operands)
    return cls(*operands)


class Unary(Item):
    """Parent class for unary operations

    :param a: the expression the operator applies on
    """
    _children_attrs = ["a"]

    def __init__(self, a, **kwargs):
        self.a = a
        super().__init__(**kwargs)

    def __str__(self, head_tail=False):
        value = "%s%s" % (self.op, self.a.__str__(head_tail=True))
        return self._head_tail(value, head_tail)


class Plus(Unary):
    """plus, unary operation
    """
    op = "+"


class Not(Unary):
    op = 'NOT'


class Prohibit(Unary):
    """The negation
    """
    op = "-"



"""Analyzing elasticSearch schema to provide helpers for query transformation
"""


class SchemaAnalyzer:
    """An helper that analyze ElasticSearch schema, to give you suitable options
    to use when transforming queries.

    :param dict schema: the index settings as a dict.
    """

    def __init__(self, schema):
        self.settings = schema.get("settings", {})
        mappings = schema.get("mappings", {})
        if mappings.get("properties"):
            # ES >= 6 : one document type per index
            self.mappings = {"_doc": mappings}
        else:
            # ES < 6 : multiple document types per index allowed
            self.mappings = mappings

    def _dot_name(self, fname, parents):
        return ".".join([p[0] for p in parents] + [fname])

    def default_field(self):
        try:
            return self.settings["query"]["default_field"]
        except KeyError:
            return "*"

    def _walk_properties(self, properties, parents=None, subfields=False):
        if parents is None:
            parents = []
        for fname, fdef in properties.items():
            yield fname, fdef, parents
            if subfields and "fields" in fdef:
                subfield_parents = parents + [(fname, fdef)]
                subdef = dict(fdef)  # sub field definition overload their parents one
                subfield_defs = subdef.pop("fields")
                for fname, fdef in subfield_defs.items():
                    fdef = dict(subdef, **fdef)
                    yield fname, fdef, subfield_parents
            inner_properties = fdef.get("properties", {})
            if inner_properties:
                new_parents = parents + [(fname, fdef)]
                yield from self._walk_properties(inner_properties, new_parents, subfields)

    def iter_fields(self, subfields=False):
        for mapping in self.mappings.values():
            yield from self._walk_properties(mapping.get("properties", {}), subfields=subfields)

    def not_analyzed_fields(self):
        for fname, fdef, parents in self.iter_fields(subfields=True):
            not_analyzed = (
                (fdef.get("type") == "string" and fdef.get("index", "") == "not_analyzed") or
                fdef.get("type") not in ("text", "string", "nested", "object")
            )
            if not_analyzed:
                yield self._dot_name(fname, parents)

    def nested_fields(self):
        result = {}
        for fname, fdef, parents in self.iter_fields():
            pdef = parents[-1][1] if parents else {}
            if pdef.get("type") == "nested":
                target = result
                cumulated = []
                for n, _ in parents:
                    cumulated.append(n)
                    key = ".".join(cumulated)
                    if key in target:
                        target = target[key]
                        cumulated = []
                if cumulated:
                    key = ".".join(cumulated)
                    target = target.setdefault(key, {})
                target[fname] = {}
        return result

    def object_fields(self):
        for fname, fdef, parents in self.iter_fields():
            pdef = parents[-1][1] if parents else {}
            if pdef.get("type") == "object" and fdef.get("type") not in ("object", "nested"):
                yield self._dot_name(fname, parents)

    def sub_fields(self):
        """return all known subfields
        """
        # we do not ask subfields, for they would be lost in the mass
        for fname, fdef, parents in self.iter_fields():
            subfields = fdef.get("fields")
            if subfields:
                subfield_parents = parents + [(fname, fdef)]
                for subname in subfields:
                    yield self._dot_name(subname, subfield_parents)

    def query_builder_options(self):
        """return options suitable for
        :py:class:`luqum.elasticsearch.visitor.ElasticsearchQueryBuilder`
        """
        return {
            "default_field": self.default_field(),
            "not_analyzed_fields": list(self.not_analyzed_fields()),
            "nested_fields": self.nested_fields(),
            "object_fields": list(self.object_fields()),
        }



import abc
import re

from ..tree import Term


class JsonSerializableMixin:
    """
    Mixin to force subclasses to implement the json method
    """

    @property
    @abc.abstractmethod
    def json(self):
        pass  # pragma: no cover


class AbstractEItem(JsonSerializableMixin):
    """
    Base item element to build the "item" json
    For instance : {"term": {"field": {"value": "query"}}}
    """

    boost = None
    _fuzzy = None

    _KEYS_TO_ADD = ('boost', 'fuzziness', '_name')
    ADDITIONAL_KEYS_TO_ADD = ()

    def __init__(self, no_analyze=None, method='term', fields=[], _name=None, field_options=None):
        self._method = method
        self._fields = fields
        self._no_analyze = no_analyze if no_analyze else []
        self.zero_terms_query = 'none'
        self.field_options = field_options or {}
        if _name is not None:
            self._name = _name

    @property
    def json(self):
        field = self.field
        inner_json = dict(self.field_options.get(field, {}))
        result = inner_json.pop('match_type', None)  # remove "match_type" key
        if not result:  # conditionally remove "type" (for backward compatibility)
            inner_json.pop('type', None)
        if self.method in ['query_string', 'multi_match']:
            json = {self.method: inner_json}
        else:
            json = {self.method: {field: inner_json}}

        # add base conf
        keys = self._KEYS_TO_ADD + self.ADDITIONAL_KEYS_TO_ADD
        for key in keys:
            value = getattr(self, key, None)
            if value is not None:
                if key == 'q':
                    if 'match' in self.method:
                        inner_json['query'] = value
                        if self.method == 'match':
                            inner_json['zero_terms_query'] = self.zero_terms_query
                    elif self.method == 'query_string':
                        inner_json['query'] = value
                        inner_json['default_field'] = self.field
                        inner_json['analyze_wildcard'] = inner_json.get('analyze_wildcard', True)
                        inner_json['allow_leading_wildcard'] = inner_json.get(
                            'allow_leading_wildcard', True)
                    else:
                        inner_json['value'] = value
                else:
                    inner_json[key] = value
        return json

    @property
    def field(self):
        return '.'.join(self._fields)

    @property
    def fuzziness(self):
        return self._fuzzy

    @fuzziness.setter
    def fuzziness(self, fuzzy):
        self._method = 'fuzzy'
        self._fuzzy = fuzzy

    def _value_has_wildcard_char(self):
        # reuse the work done in Term
        term = Term(getattr(self, 'q', ''))
        return term.has_wildcard()

    def _is_analyzed(self):
        return self.field not in self._no_analyze

    @property
    def method(self):
        is_analyzed = self._is_analyzed()
        if not is_analyzed and self._value_has_wildcard_char():
            return 'wildcard'
        elif is_analyzed:
            if self._value_has_wildcard_char():
                return 'query_string'
            elif self._method.startswith("match"):
                options = self.field_options.get(self.field, {})
                # Support the type opiton for backward compatibility
                return options.get("match_type", options.get("type", self._method))
        return self._method


class EWord(AbstractEItem):
    """
    Build a word

    ::
        >>> from unittest import TestCase
        >>> TestCase().assertDictEqual(
        ...     EWord(q='test', fields=["text"]).json,
        ...     {'term': {'text': {
        ...         'value': 'test'
        ...     }}},
        ... )
    """

    ADDITIONAL_KEYS_TO_ADD = ('q', )

    def __init__(self, q, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.q = q

    @property
    def json(self):
        # field:* is transformed to exists query
        if self.q == '*':
            query = {"exists": {"field": self.field}}
            name = getattr(self, "_name", None)
            if name is not None:
                query["exists"]["_name"] = name
            return query
        return super().json


class EPhrase(AbstractEItem):
    """
    Build a phrase

    ::
        >>> from unittest import TestCase
        >>> TestCase().assertDictEqual(
        ...     EPhrase(phrase='"another test"', fields=["text"]).json,
        ...     {'match_phrase': {'text': {'query': 'another test'}}},
        ... )
    """

    ADDITIONAL_KEYS_TO_ADD = ('q',)
    _proximity = None

    def __init__(self, phrase, *args, **kwargs):
        super().__init__(method='match_phrase', *args, **kwargs)
        phrase = self._replace_CR_and_LF_by_a_whitespace(phrase)
        self.q = self._remove_double_quotes(phrase)

    def __repr__(self):
        return "%s(%s=%s)" % (self.__class__.__name__, self.field, self.q)

    def _replace_CR_and_LF_by_a_whitespace(self, phrase):
        return re.sub(r'\s+', ' ', phrase)

    def _remove_double_quotes(self, phrase):
        return phrase[1:-1]

    def _value_has_wildcard_char(self):
        # Wildcard not active in Phrase
        return False

    @property
    def slop(self):
        return self._proximity

    @slop.setter
    def slop(self, slop):
        self._proximity = slop
        self.ADDITIONAL_KEYS_TO_ADD += ('slop', )


class ERange(AbstractEItem):
    """
    Build a range
    ::

        >>> from unittest import TestCase
        >>> TestCase().assertDictEqual(
        ...     ERange(lt=100, gte=10, fields=["text"]).json,
        ...     {'range': {'text': {'lt': 100, 'gte': 10}}},
        ... )
    """

    def __init__(self, lt=None, lte=None, gt=None, gte=None, *args, **kwargs):
        super().__init__(method='range', *args, **kwargs)
        if lt and lt != '*':
            self.lt = lt
            self.ADDITIONAL_KEYS_TO_ADD += ('lt', )
        elif lte and lte != '*':
            self.lte = lte
            self.ADDITIONAL_KEYS_TO_ADD += ('lte', )
        if gt and gt != '*':
            self.gt = gt
            self.ADDITIONAL_KEYS_TO_ADD += ('gt', )
        elif gte and gte != '*':
            self.gte = gte
            self.ADDITIONAL_KEYS_TO_ADD += ('gte', )


class AbstractEOperation(JsonSerializableMixin):
    pass


class EOperation(AbstractEOperation):
    """
    Abstract operation taking care of the json build
    """

    def __init__(self, items, **options):
        self.items = items
        self._method = None
        self.options = options

    def __repr__(self):
        items = ", ".join(i.__repr__() for i in self.items)
        return "%s(%s)" % (self.__class__.__name__, items)

    @property
    def json(self):
        bool_query = {self.operation: [item.json for item in self.items]}
        query = dict(bool_query, **self.options)
        return {'bool': query}


class ENested(AbstractEOperation):
    """
    Build ENested element

    Take care to remove ENested children
    """

    def __init__(self, nested_path, nested_fields, items, *args, _name=None, **kwargs):

        self._nested_path = [nested_path]
        self.items = self._exclude_nested_children(items)
        self._name = _name

    @property
    def nested_path(self):
        return '.'.join(self._nested_path)

    def __repr__(self):
        return "%s(%s, %s)" % (self.__class__.__name__, self.nested_path, self.items)

    def _exclude_nested_children(self, subtree):
        """
        Rebuild tree excluding ENested in children if some are present

        ::
            >>> from unittest import TestCase
            >>> tree = EMust(items=[
            ...     ENested(
            ...         nested_path='a',
            ...         nested_fields=['a'],
            ...         items=EPhrase('"François"', fields=["text"])
            ...     ),
            ...     ENested(
            ...         nested_path='a',
            ...         nested_fields=['a'],
            ...         items=EPhrase('"Dupont"', fields=["text"]))
            ... ])
            >>> nested_node = ENested(
            ...     nested_path='a', nested_fields=['a'], items=tree)
            >>> TestCase().assertEqual(
            ...     nested_node.__repr__(),
            ...     'ENested(a, EMust(EPhrase(text=François), EPhrase(text=Dupont)))'
            ... )
        """
        if isinstance(subtree, ENested):
            # Exclude ENested

            if subtree.nested_path == self.nested_path:
                return self._exclude_nested_children(subtree.items)
            else:
                return subtree
        elif isinstance(subtree, AbstractEOperation):
            # Exclude ENested in children
            subtree.items = [
                self._exclude_nested_children(child)
                for child in subtree.items
            ]
            return subtree
        else:
            # return the subtree once ENested has been excluded
            return subtree

    @property
    def json(self):
        data = {'nested': {'path': self.nested_path, 'query': self.items.json}}
        if self._name:
            data['nested']['_name'] = self._name
        return data


class EShould(EOperation):
    """
    Build a should operation

    ::
        >>> from unittest import TestCase
        >>> json = EShould(
        ...     items=[EPhrase('"monty python"', fields=["text"]),
        ...            EPhrase('"spam eggs"', fields=["text"])]
        ... ).json
        >>> TestCase().assertDictEqual(
        ...     json,
        ...     {'bool': {'should': [
        ...         {'match_phrase': {'text':
        ...             {'query': 'monty python'}}},
        ...         {'match_phrase': {'text': {'query': 'spam eggs'}}},
        ...     ]}}
        ... )
    """
    operation = 'should'


class AbstractEMustOperation(EOperation):

    def __init__(self, items, **options):
        op = super().__init__(items, **options)
        for item in self.items:
            item.zero_terms_query = self.zero_terms_query
        return op


class EMust(AbstractEMustOperation):
    """
    Build a must operation

    ::
        >>> from unittest import TestCase
        >>> json = EMust(
        ...     items=[EPhrase('"monty python"', fields=["text"]),
        ...            EPhrase('"spam eggs"', fields=["text"])]
        ... ).json
        >>> TestCase().assertDictEqual(
        ...     json,
        ...     {'bool': {'must': [
        ...         {'match_phrase': {'text':
        ...             {'query': 'monty python'}}},
        ...         {'match_phrase': {'text':
        ...             {'query': 'spam eggs'}}},
        ...     ]}}
        ... )
    """
    zero_terms_query = 'all'
    operation = 'must'


class EMustNot(AbstractEMustOperation):
    """
    Build a must not operation

    ::
        >>> from unittest import TestCase
        >>> TestCase().assertDictEqual(
        ...     EMustNot(items=[EPhrase('"monty python"', fields=["text"])],).json,
        ...     {'bool': {'must_not': [
        ...         {'match_phrase': {'text':
        ...             {'query': 'monty python'}}},
        ...     ]}}
        ... )
    """
    zero_terms_query = 'none'
    operation = 'must_not'


class EBoolOperation(EOperation):

    @property
    def json(self):
        must_items = []
        should_items = []
        must_not_items = []
        for item in self.items:
            if isinstance(item, EMust):
                must_items.extend(item.items)
            elif isinstance(item, EMustNot):
                must_not_items.extend(item.items)
            else:
                should_items.append(item)
        bool_query = {}
        if must_items:
            bool_query["must"] = [item.json for item in must_items]
        if should_items:
            bool_query["should"] = [item.json for item in should_items]
        if must_not_items:
            bool_query["must_not"] = [item.json for item in must_not_items]

        query = dict(bool_query, **self.options)
        return {'bool': query}


class ElasticSearchItemFactory:
    """
    Factory to preconfigure EItems and EOperation

    ::
        >>> from unittest import TestCase
        >>> factory = ElasticSearchItemFactory(
        ...     no_analyze=['text'], nested_fields=[], field_options={"text": {"slop": 1}})
        >>> word = factory.build(EWord, q='test', fields=["text"])
        >>> TestCase().assertDictEqual(
        ...     word.json,
        ...     {'term': {'text': {'value': 'test', "slop": 1}}},
        ... )
    """

    def __init__(self, no_analyze, nested_fields, field_options):
        self._no_analyze = no_analyze
        self._nested_fields = nested_fields
        self._nested_fields = nested_fields
        self._field_options = field_options

    def build(self, cls, *args, **kwargs):
        # add parameters based on item type
        if issubclass(cls, AbstractEItem):
            # eventually add field defaults to kwargs
            if "field_options" not in kwargs:
                kwargs = dict(kwargs, field_options=self._field_options)
            return cls(
                no_analyze=self._no_analyze,
                *args,
                **kwargs
            )
        elif cls is ENested:
            return cls(
                nested_fields=self._nested_fields,
                *args,
                **kwargs
            )
        else:
            return cls(*args, **kwargs)



"""Support for naming expressions

In order to use elastic search named query, we need to be able to assign names to expressions
and retrieve their positions in the query text.

This module adds support for that.
"""
from . import tree
from .visitor import PathTrackingVisitor, PathTrackingTransformer


#: Names are added to tree items via an attribute named `_luqum_name`
NAME_ATTR = "_luqum_name"


def set_name(node, value):
    setattr(node, NAME_ATTR, value)


def get_name(node):
    return getattr(node, NAME_ATTR, None)


class TreeAutoNamer(PathTrackingVisitor):
    """Helper for :py:func:`auto_name`
    """

    LETTERS = "abcdefghilklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    _pos_letter = {l: i for i, l in enumerate(LETTERS)}

    def next_name(self, name):
        """Given name, return next name

        ::
           >>> tan = TreeAutoNamer()
           >>> tan.next_name(None)
           'a'
           >>> tan.next_name('aZ')
           'aZa'
           >>> tan.next_name('azb')
           'azc'
        """
        if name is None:
            # bootstrap
            return self.LETTERS[0]
        else:
            actual_pos = self._pos_letter[name[-1]]
            try:
                # we want to increment last letter
                return name[:-1] + self.LETTERS[actual_pos + 1]
            except IndexError:
                # we exhausts letters, add a new one instead
                return name + self.LETTERS[0]

    def visit_base_operation(self, node, context):
        """name is to be set on children of operations
        """
        # put a _name on each children
        name = context["global"]["name"]
        for i, child in enumerate(node.children):
            name = self.next_name(name)
            set_name(child, name)
            # remember name to path
            context["global"]["name_to_path"][name] = context["path"] + (i,)
        # put name back in global context
        context["global"]["name"] = name
        yield from self.generic_visit(node, context)

    def visit(self, node):
        """visit the tree and add names to nodes while tracking their path
        """
        # trick: we use a "global" dict inside context dict so that when we copy context,
        # we still track the same objects
        context = {"global": {"name": None, "name_to_path": {}}}
        super().visit(node, context)
        name_to_path = context["global"]["name_to_path"]
        # handle special case, if we have no name so far, put one on the root
        if not name_to_path:
            node_name = self.next_name(context["global"]["name"])
            set_name(node, node_name)
            name_to_path[node_name] = ()
        return name_to_path


def auto_name(tree, targets=None, all_names=False):
    """Automatically add names to nodes of a parse tree, in order to be able to track matching.

    We add them to top nodes under operations as this is where it is useful for ES named queries

    :return dict: association of name with the path (as a tuple) to a the corresponding children
    """
    return TreeAutoNamer().visit(tree)


def matching_from_names(names, name_to_path):
    """Utility to convert a list of name and the result of auto_name
    to the matching parameter for :py:class:`MatchingPropagator`

    :param list names: list of names
    :param dict name_to_path: association of names with path to children
    :return tuple: (set of matching paths, set of other known paths)
    """
    matching = {name_to_path[name] for name in names}
    return (matching, set(name_to_path.values()) - matching)


def element_from_path(tree, path):
    """Given a tree, retrieve element corresponding to path

    :param luqum.tree.Item tree: luqum expression tree
    :param tuple path: tuple representing top down access to a child
    :return  luqum.tree.Item: target item
    """
    # python likes iterations over recursivity
    node = tree
    path = list(path)
    while path:
        node = node.children[path.pop(0)]
    return node


def element_from_name(tree, name, name_to_path):
    return element_from_path(tree, name_to_path[name])


class MatchingPropagator:
    """Class propagating matching to upper elements based on known base element matching

    :param luqum.tree.Item default_operation: tells how to treat UnknownOperation.
        Choose between :py:class:`luqum.tree.OrOperation` and :py:class:`luqum.tree.AndOperation`
    """

    OR_NODES = (tree.OrOperation,)
    """A tuple of nodes types considered as OR operations
    """
    NEGATION_NODES = (tree.Not, tree.Prohibit)
    """A tuple of nodes types considered as NOT operations
    """
    NO_CHILDREN_PROPAGATE = (tree.Range, tree.BaseApprox)
    """A tuple of nodes for which propagation is of no use
    """

    def __init__(self, default_operation=tree.OrOperation):
        if default_operation is tree.OrOperation:
            self.OR_NODES = self.OR_NODES + (tree.UnknownOperation,)

    def _status_from_parent(self, path, matching, other):
        """Get status from nearest parent in hierarchie which had a name
        """
        if path in matching:
            return True
        elif path in other:
            return False
        elif not path:
            return False
        else:
            return self._status_from_parent(path[:-1], matching, other)

    def _propagate(self, node, matching, other, path):
        """recursively propagate matching

        return tuple: (
            node is matching,
            set of pathes of matching sub nodes,
            set of pathes of non matching sub nodes)
        """
        paths_ok = set()  # path of nodes that are matching
        paths_ko = set()  # path of nodes that are not matching
        children_status = []  # bool for each children, indicating if it matches or not
        # recurse children
        if node.children and not isinstance(node, self.NO_CHILDREN_PROPAGATE):
            for i, child in enumerate(node.children):
                child_ok, sub_ok, sub_ko = self._propagate(
                    child, matching, other, path + (i,),
                )
                paths_ok.update(sub_ok)
                paths_ko.update(sub_ko)
                children_status.append(child_ok)
        # resolve node status
        if path in matching:
            node_ok = True
        elif children_status:  # compute from children
            # compute parent success from children
            operator = any if isinstance(node, self.OR_NODES) else all
            node_ok = operator(children_status)
        else:
            node_ok = self._status_from_parent(path, matching, other)
        if isinstance(node, self.NEGATION_NODES):
            # negate result
            node_ok = not node_ok
        # add node to the right set
        target_set = paths_ok if node_ok else paths_ko
        target_set.add(path)
        # return result
        return node_ok, paths_ok, paths_ko

    def __call__(self, tree, matching, other=frozenset()):
        """
        Given a list of paths that are known to match,
        return all pathes in the tree that are matches.

        .. note:: we do not descend into nodes that are positive.
           Normally matching just provides nodes at the right levels
           for propagation to be effective.
           Descending would mean risking to give non consistent information.

        :param list matching: list of path of matching nodes (each path is a tuple)
        :param list other: list of other path that had a name, but were not reported as matching

        :return tuple: (
            set of matching path after propagation,
            set of non matching pathes after propagation)
        """
        tree_ok, paths_ok, paths_ko = self._propagate(tree, matching, other, ())
        return paths_ok, paths_ko


class ExpressionMarker(PathTrackingTransformer):
    """A visitor to mark a tree based on elements belonging to a path or not

    One intended usage is to add marker around nodes matching a request,
    by altering tail and head of elements
    """

    def mark_node(self, node, path, *info):
        """implement this in your own code, maybe altering the head / tail arguments
        """
        return node

    def generic_visit(self, node, context):
        # we simply generate new_node and mark it
        new_node, = super().generic_visit(node, context)
        yield self.mark_node(new_node, context["path"], *context["info"])

    def __call__(self, tree, *info):
        return self.visit(tree, context={"info": info})


class HTMLMarker(ExpressionMarker):
    """from paths that are ok or ko, add html elements with right class around elements

    :param str ok_class: class for elements in paths_ok
    :param str ko_class: class for elements in paths_ko
    :param str element: html element used to surround sub expressions
    """

    def __init__(self, ok_class="ok", ko_class="ko", element="span"):
        super().__init__()
        self.ok_class = ok_class
        self.ko_class = ko_class
        self.element = element

    def css_class(self, path, paths_ok, paths_ko):
        return self.ok_class if path in paths_ok else self.ko_class if path in paths_ko else None

    def mark_node(self, node, path, paths_ok, paths_ko, parcimonious):
        node_class = self.css_class(path, paths_ok, paths_ko)
        add_class = node_class is not None
        if add_class and parcimonious:
            # find nearest parent with a class
            parent_class = None
            parent_path = path
            while parent_class is None and parent_path:
                parent_path = parent_path[:-1]
                parent_class = self.css_class(parent_path, paths_ok, paths_ko)
            # only add class if different from parent
            add_class = node_class != parent_class
        if add_class:
            node.head = f'<{self.element} class="{node_class}">{node.head}'
            node.tail = f'{node.tail}</{self.element}>'
        return node

    def __call__(self, tree, paths_ok, paths_ko, parcimonious=True):
        """representation of tree, adding html elements with right class around subexpressions
        according to their presence in paths_ok or paths_ko

        :param tree: a luqum tree
        :param paths_ok: set of path to nodes (express as tuple of int) that should get ok_class
        :param paths_ko: set of path to nodes that should get ko_class
        :param parcimonious: only add class when parent node does not have same class

        :return str: expression with html elements surrounding part of expression
          with right class attribute according to paths_ok and paths_ko
        """
        new_tree = super().__call__(tree, paths_ok, paths_ko, parcimonious)
        return new_tree.__str__(head_tail=True)



"""If you have a query with a nested query containing operations,
when using named queries, Elasticsearch won't report inner matching.

This is a problem if you extensively use it.
"""


def get_first_name(query):
    if isinstance(query, dict):
        if "_name" in query:
            return query["_name"]
        elif "bool" in query:
            # do not go down bool
            return None
        else:
            children = query.values()
    elif isinstance(query, list):
        children = query
    else:
        return None
    iter_candidates = (get_first_name(child) for child in children)
    candidates = [candidate for candidate in iter_candidates if candidate is not None]
    return candidates[0] if candidates else None


def extract_nested_queries(query, query_nester=None):
    """given a query,
    extract all queries that are under a nested query and boolean operations,
    returning an atomic nested version of them.
    Those nested queries, also take care of changing the name to the nearest inner name,

    This is useful for Elasticsearch won't go down explaining why a nested query is matching.

    :param dict query: elasticsearch query to analyze
    :param callable query_nester: this is the function called to nest sub queries, leave it default
    :return list: queries that you should run to get all matching

    .. note:: because we re-nest part of bool queries, results might not be accurate
       for::
          {"bool": "must" : [
              {"nested": {"path": "a", "match": {"x": "y"}}},
              {"nested": {"path": "a", "match": {"x": "z"}}}
          ]}
       is not the same as::
          {"nested": {"path": "a", "bool": "must": [{"match": {"x": "y"}}, {"match": {"x": "z"}}]}}

       if x is multivalued.
       The first would match `{"a": [{"x": "y"}, {"x": "z"}]}`
       While the second would only match if `x` contains `"y z"` or `"z y"`
    """
    queries = []  # this contains our result
    in_nested = query_nester is not None
    sub_query_nester = query_nester
    if isinstance(query, dict):
        if "nested" in query:
            params = {k: v for k, v in query["nested"].items() if k not in ("query", "name")}

            def sub_query_nester(req, name):
                nested = {"nested": {"query": req, **params}}
                if query_nester is not None:
                    nested = query_nester(nested, name)
                if name is not None:
                    nested["nested"]["_name"] = name
                return nested

        bool_param = {"must", "should", "must_not"} & set(query.keys())
        if bool_param and in_nested:
            # we are in a list of operations in a bool inside a nested,
            # make a query with nested on sub arguments
            op, = bool_param  # must or should or must_not
            # normalize to a list
            sub_queries = query[op] if isinstance(query[op], list) else [query[op]]
            # add nesting
            nested_sub_queries = [
                query_nester(sub_query, get_first_name(sub_query)) for sub_query in sub_queries
            ]
            # those are queries we want to return
            queries.extend(nested_sub_queries)
            # continue processing in each sub query
            # (before nesting, nesting is contained in query_nester)
            children = sub_queries
        else:
            children = query.values()
    elif isinstance(query, list):
        children = query
    else:
        # leaf: final recursivity
        children = []

    # recurse
    for child_query in children:
        queries.extend(
            extract_nested_queries(child_query, query_nester=sub_query_nester)
        )
    return queries



import threading

from ply import lex

from . import parser


thread_local = threading.local()


def parse(input=None, lexer=None, debug=False, tracking=False):
    """A (hopefully) thread safe version of :py:meth:`luqum.parser.parse`

    PLY is not thread safe because of its lexer state, but cloning it we can be thread safe.
    see: https://github.com/jurismarches/luqum/issues/72
    """
    if not hasattr(thread_local, "lexer"):
        thread_local.lexer = lex.lexer.clone()
    return parser.parser.parse(input, lexer=thread_local.lexer)



# -*- coding: utf-8 -*-
"""Base classes to implement a visitor pattern.
"""


def camel_to_lower(name):
    return "".join(
        "_" + w.lower() if w.isupper() else w.lower()
        for w in name).lstrip("_")


class TreeVisitor:
    """
    Tree Visitor base class.

    This class is meant to be subclassed, with the subclass implementing
    visitor methods for each Node type it is interested in.

    By default, those visitor method should be named ``'visit_'`` + class
    name of the node, converted to lower_case (ie: visit_search_node for a
    SearchNode class)[#tweakvisit]_.

    It's up to the visit method of each node to recursively call children (or not)
    It may be done simply by calling the generic_visit method.

    By default the `generic_visit`, simply trigger visit of subnodes, yielding no information.

    If the goal is to modify the initial tree, to get a new modified copy
    use :py:class:`TreeTranformer` instead.

    .. [#tweakvisit]: You can tweak this behaviour
       by overriding the `visitor_method_prefix` & `generic_visitor_method_name` class attributes.

    :param bool track_parents: if True the context will contain parents of current node as a list.
        It's up to you to maintain this list in your own methods.
    """
    visitor_method_prefix = 'visit_'
    generic_visitor_method_name = 'generic_visit'

    def __init__(self, track_parents=False):
        self.track_parents = track_parents

    _get_method_cache = None

    def _get_method(self, node):
        if self._get_method_cache is None:
            self._get_method_cache = {}
        try:
            meth = self._get_method_cache[type(node)]
        except KeyError:
            for cls in node.__class__.mro():
                try:
                    method_name = "{}{}".format(
                        self.visitor_method_prefix,
                        camel_to_lower(cls.__name__)
                    )
                    meth = getattr(self, method_name)
                    break
                except AttributeError:
                    continue
            else:
                meth = getattr(self, self.generic_visitor_method_name)
            self._get_method_cache[type(node)] = meth
        return meth

    def visit(self, tree, context=None):
        """Traversal of tree

        :param luqum.tree.Item tree: a tree representing a lucene expression
        :param dict context: a dict with initial values for context

        .. note:: the values in context, are not guaranteed to move up the hierachy,
           because we do copy of context for children to have specific values.

           A trick you can use if you need values to move up the hierachy
           is to set a `"global"` key containing a dict, where you can store values.
        """
        if context is None:
            context = {}
        return list(self.visit_iter(tree, context=context))

    def visit_iter(self, node, context):
        """
        Basic, recursive traversal of the tree.

        :param list parents: the list of parents
        :param dict context: a dict of contextual variable for free use
            to track states while traversing the tree (eg. the current field name)
        """
        method = self._get_method(node)
        yield from method(node, context)

    def child_context(self, node, child, context, **kwargs):
        """Generate a context for children.

        The context children is distinct from its parent context,
        so that visit in a branch does not affect others.

        .. note:: If you need global parameters,
            a trick is to put them in dict in a "global" entry
            as we do a swallow copy of context, and not a deep one.

        :param luqum.tree.Item node: parent node
        :param luqum.tree.Item child: child node
        :param dict context: parent context
        :return dict: child context
        """
        child_context = dict(context)
        if self.track_parents:
            child_context["parents"] = context.get("parents", ()) + (node,)
        return child_context

    def generic_visit(self, node, context):
        """
        Default visitor function, called if nothing matches the current node.

        It simply visit children.

        :param luqum.tree.Item node: current node
        :param dict context: context (aka local parameters received from parents)
        """
        for child in node.children:
            child_context = self.child_context(node, child, context)
            yield from self.visit_iter(child, context=child_context)


class TreeTransformer(TreeVisitor):
    """A version of TreeVisitor that is aimed at obtaining a transformed copy of tree.

    .. note:: It is far better to build a transformed copy,
        than to modify in place the original tree, as it is less error prone.

    :param bool track_new_parents: do we want to track new parents in the context ?
    """

    def __init__(self, track_new_parents=False, **kwargs):
        self.track_new_parents = track_new_parents
        super().__init__(**kwargs)

    def _clone_item(self, node):
        """simply call node.clone_item

        Surcharge this method to add specific tweaks if needed (like copying special attributes)
        """
        return node.clone_item()

    def visit(self, tree, context=None):
        """Visit the tree, by default building a copy and returning it.

        :param luqum.tree.Item tree: luqum expression tree
        :param context: optional initial context
        """
        if context is None:
            context = {}
        try:
            value, = self.visit_iter(tree, context=context)
            return value
        except ValueError as e:
            if str(e).startswith(("too many values to unpack", "not enough values to unpack")):
                exc = ValueError(
                    "The visit of the tree should have produced exactly one element "
                    "(the transformed tree)"
                )
                raise exc from e
            else:
                raise

    def child_context(self, node, child, context, **kwargs):
        child_context = super().child_context(node, child, context, **kwargs)
        if self.track_new_parents:
            child_context["new_parents"] = context.get("new_parents", ()) + (kwargs["new_node"],)
        return child_context

    def generic_visit(self, node, context):
        """
        Default visitor function, called if nothing matches the current node.

        It simply clone node and children
        """
        new_node = self._clone_item(node)
        new_node.children = list(self.clone_children(node, new_node, context))
        yield new_node

    def clone_children(self, node, new_node, context):
        """Helper to clone children.

        .. note:: a children may generate more than one children or none, for flexibility
           but it's up to the transformer to ensure everything is ok
        """
        for child in node.children:
            child_context = self.child_context(node, child, context, new_node=new_node)
            new_children = self.visit_iter(child, context=child_context)
            for new_child in new_children:
                yield new_child


class PathTrackingMixin:
    """It can be useful to compute path of an element (as tuple of index in parent children)

    This mixin provides base components
    """

    def child_context(self, node, child, context, **kwargs):
        """Thanks to "path" and "position" in kwargs, we add the path of children
        """
        child_context = super().child_context(node, child, context, **kwargs)
        child_context["path"] = context["path"] + (kwargs["position"],)
        return child_context

    def visit(self, node, context=None):
        """visit the tree while tracking their path
        """
        if context is None:
            context = {}
        context["path"] = ()
        return super().visit(node, context=context)


class PathTrackingVisitor(PathTrackingMixin, TreeVisitor):
    """Path tracking version of TreeVisitor
    """

    def generic_visit(self, node, context):
        for i, child in enumerate(node.children):
            child_context = self.child_context(node, child, context, position=i)
            yield from self.visit_iter(child, context=child_context)


class PathTrackingTransformer(PathTrackingMixin, TreeTransformer):
    """Path tracking version of TreeTransformer
    """

    def clone_children(self, node, new_node, context):
        for i, child in enumerate(node.children):
            child_context = self.child_context(node, child, context, new_node=new_node, position=i)
            new_children = self.visit_iter(child, context=child_context)
            for new_child in new_children:
                yield new_child
